package net.jakewilliamson.sentry;

import android.app.Notification;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.net.VpnService;
import android.os.ParcelFileDescriptor;
import android.preference.PreferenceManager;
import android.support.v4.content.LocalBroadcastManager;
import android.util.Log;

import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import static net.jakewilliamson.sentry.Util.closeResources;

public class LocalVPNService extends VpnService
{
    public static final String BROADCAST_VPN_STATE = "net.jakewilliamson.sentry.VPN_STATE";
    private static final String TAG = LocalVPNService.class.getSimpleName();
    private static final String VPN_ADDRESS = "10.0.0.2"; // Only IPv4 support for now
    private static final String VPN_ROUTE = "0.0.0.0"; // Intercept everything
    private static final Integer NOTIFICATION_COUNT = 15; // Block notifications of this type for 10 attempts
    private static boolean isRunning = false;
    private static final List<String> hintList = new ArrayList<>();
    private static final List<String> valueList = new ArrayList<>();
    private static boolean hasData;
    private static Context context;
    private static final HashMap<String, Integer> sleepList = new HashMap<>();
    private ParcelFileDescriptor vpnInterface = null;
    private PendingIntent pendingIntent;
    private ConcurrentLinkedQueue<Packet> deviceToNetworkUDPQueue;
    private ConcurrentLinkedQueue<Packet> deviceToNetworkTCPQueue;
    private ConcurrentLinkedQueue<ByteBuffer> networkToDeviceQueue;
    private ExecutorService executorService;
    private Selector udpSelector;
    private Selector tcpSelector;

    public static Context getPublicContext()
    {
        return context;
    }

    public static boolean isRunning()
    {
        return isRunning;
    }

    @Override
    public void onCreate()
    {
        super.onCreate();
        context = getBaseContext();
        isRunning = true;
        setupVPN();
        hasData = getData();

        try
        {
            udpSelector = Selector.open();
            tcpSelector = Selector.open();
            deviceToNetworkUDPQueue = new ConcurrentLinkedQueue<>();
            deviceToNetworkTCPQueue = new ConcurrentLinkedQueue<>();
            networkToDeviceQueue = new ConcurrentLinkedQueue<>();

            executorService = Executors.newFixedThreadPool(5);
            executorService.submit(new UDPInput(networkToDeviceQueue, udpSelector));
            executorService.submit(new UDPOutput(deviceToNetworkUDPQueue, udpSelector, this));
            executorService.submit(new TCPInput(networkToDeviceQueue, tcpSelector));
            executorService.submit(new TCPOutput(deviceToNetworkTCPQueue, networkToDeviceQueue, tcpSelector, this));
            executorService.submit(new VPNRunnable(vpnInterface.getFileDescriptor(),
                    deviceToNetworkUDPQueue, deviceToNetworkTCPQueue, networkToDeviceQueue));
            LocalBroadcastManager.getInstance(this).sendBroadcast(new Intent(BROADCAST_VPN_STATE).putExtra("running", true));
            Log.i(TAG, "Started");
        } catch (IOException e)
        {
            Log.e(TAG, "Error starting service", e);
            cleanup();
        }
    }

    private boolean getData()
    {
        final String PREF_NAME = this.getString(R.string.app_name) + "_pref";
        SharedPreferences pref = this.getSharedPreferences(PREF_NAME, Context.MODE_PRIVATE);

        HashMap kvs = (HashMap) pref.getAll();

        for (Object key : kvs.keySet())
        {
            hintList.add(String.valueOf(key));
            if (!String.valueOf(kvs.get(key)).equals(""))
            {
                valueList.add(String.valueOf(kvs.get(key)));
            }
        }
        if (valueList.size() > 0)
        {
            return true;
        }

        notifyNoData();
        return false;
    }

    private void notifyNoData()
    {
        NotificationUtils notificationUtils = new NotificationUtils(this);
        Notification.Builder nb = notificationUtils.getNotification(getString(R.string.notify_title), getString(R.string.notify_body),
                false);
        notificationUtils.getManager().notify(9999, nb.build());
    }

    private void setupVPN()
    {
        if (vpnInterface == null)
        {
            Builder builder = new Builder();
            builder.addAddress(VPN_ADDRESS, 32);
            builder.addRoute(VPN_ROUTE, 0);
            vpnInterface = builder.setSession(getString(R.string.app_name)).setConfigureIntent(pendingIntent).establish();
        }
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId)
    {
        return START_STICKY;
    }

    @Override
    public void onDestroy()
    {
        super.onDestroy();
        isRunning = false;
        executorService.shutdownNow();
        cleanup();
        Log.i(TAG, "Stopped");
    }

    private void cleanup()
    {
        deviceToNetworkTCPQueue = null;
        deviceToNetworkUDPQueue = null;
        networkToDeviceQueue = null;
        ByteBufferPool.clear();
        closeResources(udpSelector, tcpSelector, vpnInterface);
    }


    static class VPNRunnable implements Runnable
    {
        private static final String TAG = VPNRunnable.class.getSimpleName();
        private static final NotificationUtils notificationUtils = new NotificationUtils(context);
        private static boolean paused = false;
        private static boolean allow = true;
        private static String publicHint = "";
        private static String publicValue = "";
        private static String destIP = "";
        private static boolean lockdown = false;
        private static boolean autoBlock = false;
        private static long timeout = -1;
        private static boolean lockdownShowing = false;
        private static long foundTime;
        private final FileDescriptor vpnFileDescriptor;
        private final ConcurrentLinkedQueue<Packet> deviceToNetworkUDPQueue;
        private final ConcurrentLinkedQueue<Packet> deviceToNetworkTCPQueue;
        private final ConcurrentLinkedQueue<ByteBuffer> networkToDeviceQueue;

        VPNRunnable(FileDescriptor vpnFileDescriptor,
                    ConcurrentLinkedQueue<Packet> deviceToNetworkUDPQueue,
                    ConcurrentLinkedQueue<Packet> deviceToNetworkTCPQueue,
                    ConcurrentLinkedQueue<ByteBuffer> networkToDeviceQueue)
        {
            this.vpnFileDescriptor = vpnFileDescriptor;
            this.deviceToNetworkUDPQueue = deviceToNetworkUDPQueue;
            this.deviceToNetworkTCPQueue = deviceToNetworkTCPQueue;
            this.networkToDeviceQueue = networkToDeviceQueue;
        }

        private static void onPause()
        {
            paused = true;
        }

        public static void onResume()
        {
            paused = false;
        }

        public static void setAllow(boolean setAllow)
        {
            notificationUtils.getManager().cancel(101);
            allow = setAllow;
        }

        public static String getHint()
        {
            return publicHint;
        }

        public static String getValue()
        {
            return publicValue;
        }

        public static String getDestIP()
        {
            return destIP;
        }

        @Override
        public void run()
        {
            Log.i(TAG, "Started");

            FileChannel vpnInput = new FileInputStream(vpnFileDescriptor).getChannel();
            FileChannel vpnOutput = new FileOutputStream(vpnFileDescriptor).getChannel();

            try
            {
                ByteBuffer bufferToNetwork = null;
                boolean dataSent = true;
                boolean dataReceived;
                while (!Thread.interrupted())
                {
                    getSettings();
                    notifyLockdown();
                    if (lockdown)
                        continue;
                    if (dataSent)
                        bufferToNetwork = ByteBufferPool.acquire();
                    else
                        bufferToNetwork.clear();

                    int readBytes = vpnInput.read(bufferToNetwork);
                    if (readBytes > 0)
                    {
                        dataSent = true;
                        bufferToNetwork.flip();
                        Packet packet = new Packet(bufferToNetwork);
                        destIP = String.valueOf(packet.ip4Header.destinationAddress);
                        String data = packet.getData();

                        if (!data.equals("") && hasData)
                        {
                            checkPacket(data);
                        }

                        while (paused)
                        {
                            Thread.sleep(10);
                            checkTimer();
                        }

                        if (!allow)
                        {
                            Log.i("JAKE", "Destroying Packet");
                            packet.destroyPacket();
                            allow = true;
                        }

                        if (packet.isUDP())
                        {
                            deviceToNetworkUDPQueue.offer(packet);
                        } else if (packet.isTCP())
                        {
                            deviceToNetworkTCPQueue.offer(packet);
                        } else
                        {
                            dataSent = false;
                        }
                    } else
                    {
                        dataSent = false;
                    }

                    ByteBuffer bufferFromNetwork = networkToDeviceQueue.poll();
                    if (bufferFromNetwork != null)
                    {
                        bufferFromNetwork.flip();
                        while (bufferFromNetwork.hasRemaining())
                            vpnOutput.write(bufferFromNetwork);
                        dataReceived = true;

                        ByteBufferPool.release(bufferFromNetwork);
                    } else
                    {
                        dataReceived = false;
                    }

                    if (!dataSent && !dataReceived)
                        Thread.sleep(10);
                }
            } catch (InterruptedException e)
            {
                Log.i(TAG, "Stopping");
            } catch (IOException e)
            {
                Log.w(TAG, e.toString(), e);
            } finally
            {
                closeResources(vpnInput, vpnOutput);
            }
        }

        private void notifyLockdown()
        {

            if (lockdown && !lockdownShowing)
            {
                Notification.Builder nb = notificationUtils.showLockdown();
                notificationUtils.getManager().notify(666, nb.build());
                lockdownShowing = true;
            } else if (!lockdown && lockdownShowing)
            {
                notificationUtils.getManager().cancel(666);
                lockdownShowing = false;
            }
        }

        private void getSettings()
        {
            SharedPreferences pref = PreferenceManager.getDefaultSharedPreferences(context);
            lockdown = pref.getBoolean("androidLockdown", false);
            autoBlock = pref.getBoolean("autoAllowBlock", false);
            String timeoutString = pref.getString("Timeout", "-1");
            if (!timeoutString.equals(""))
            {
                timeout = Long.parseLong(timeoutString);
            }
        }

        private void checkPacket(String data)
        {
            String test = data.toLowerCase();

            Log.i("JAKE", test);

            for (int i = 0; i < valueList.size(); i++)
            {
                String value = valueList.get(i);

                if (test.contains(value.toLowerCase()))
                {
                    if (sleepList.containsKey(value) && sleepList.get(value) < NOTIFICATION_COUNT)
                    {
                        sleepList.put(value, sleepList.get(value) + 1);
                        break;
                    }
                    onPause();
                    startTimer();
                    if (3 < timeout || -1 == timeout)
                    {
                        notifyFound(hintList.get(i), value);
                        sleepList.put(value, 1);
                    }
                    break;
                }
            }
        }

        private void notifyFound(String hint, String value)
        {
            publicHint = hint;
            publicValue = value;
            String title = "Data Being Shared";
            String body = hint + " is currently being shared";
            Notification.Builder nb = notificationUtils.getNotification(title, body, true);
            notificationUtils.getManager().notify(101, nb.build());
        }

        private void customNotify(String title, String body)
        {
            Notification.Builder nb = notificationUtils.getNotification(title, body, false);
            notificationUtils.getManager().notify(9999, nb.build());
        }

        private void startTimer()
        {
            foundTime = System.nanoTime();
        }

        private void checkTimer()
        {
            if (0 > timeout)
            {
                return;
            }
            long currentTime = System.nanoTime();
            long elapsedTime = currentTime - foundTime;

            long timeoutNS = timeout * 1000000000;

            if (elapsedTime >= timeoutNS)
            {
                if (autoBlock)
                {
                    setAllow(false);
                    customNotify(context.getString(R.string.data_blocked_title), context.getString(R.string.data_blocked_body)
                    );
                    History.saveData(getHint(), getValue(), getDestIP(), "Blocked automatically", context);

                    Log.i("JAKE", "Blocked auto");
                } else
                {
                    setAllow(true);
                    customNotify(context.getString(R.string.data_allowed_title), context.getString(R.string.data_allowed_body)
                    );
                    History.saveData(getHint(), getValue(), getDestIP(), "Allowed automatically", context);
                    Log.i("JAKE", "Allowed auto");
                }
                onResume();
            }
        }
    }
}
